/* N-body calculation using Greenspan 1st-order discrete mechanics.	*/
/* 1-D version with vector calculation, mixed spreading and systolic	*/
/* computation.  All processors store particle information and		*/
/* calculate velocities, positions redundantly.				*/
/* Uses dynamic shape allocation to permit more flexible operation.	*/
/* This versions uses O(1) storage per vp, rather than O(NS) as in	*/
/* nb1dsys.cs, calculating the Greenspan accelerations from scratch	*/
/* on every iteration, also using a simple predictor for X1, V1.	*/
/* File:  nbigsys.cs							*/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <cscomm.h>
#include <float.h>

/* <float.h> provides useful parameters about floating-point	*/
/*  representations						*/

/* Geometry is [NM][NM][NS] with NS = systolic pipe length (even),	*/
/* NB = NM*NS,   NM*NM*NS >= physical size.				*/

int NM =  32;	/* Number of pipes in each dimension		*/
int NS =  32;	/* Number of entries in systolic pipe >= 4	*/
int SS =  16;	/* Number of systolic steps = NS/2		*/
int NB =1024;	/* Number of bodies in system			*/
int   PX;	/* Controls how much state is printed		*/
float G;	/* Gravitational constant			*/

#define TWOPI      6.2831853f

/* Particles, and pairs of particles.  All processors have particle	*/
/* states.  Some variables could be overloaded. 			*/

shape [][][]pair;	/* Will be [NM][NM][NS]	*/

/* All parallel variables are declared as pointers so 'pair' can be dynamic. */
/* State variables will be allocated as x0[3],...;  w, m, mj, r0, are floats */

/* x0, x1 - current, new positions;   v0, v1 - current, new velocities	*/
/* x0j - x0 for initial "j" particle					*/
/* m - mass;		mi,mj - scaled mass;	ac - acceleration	*/
/* w, wv, ax - working							*/
/* xm0 - conveyor (moving) positions; 3 entries	       \		*/
/* xm1 - conveyor (moving) positions; 3 entries		\ vector of	*/
/* acm - conveyor (moving) accelerations; 3 entries	/ 10 entries	*/
/* mmj - conveyor (moving) scaled masses; 1 entry      /		*/

float:pair *x0, *x1, *v0, *v1, *m, *mi, *mj, *ac, *ax, *w, *wv,
	   *x0j, *xm0, *xm1, *acm, *mmj;

/* i - row number;    j - column number;    k - pipe position	*/

int:pair *i, *j, *k;

/* diag - processor is on diagonal			*/
/* act  - processor is active pair (off-diag)		*/

bool:pair *diag, *act;

/* Data sizes for allocation, communication */

int fsz, isz, bsz, vsz, cvsz;

/* Allocate global parallel variables, pair must have been allocated	*/
/* Calculates sizes.							*/

void create_globals()
 { fsz = boolsizeof(float:pair);
   vsz  = 3*fsz;  	/* vectors	  */
   cvsz = 10*fsz;	/* the conveyor	  */
   isz  = boolsizeof(int:pair);
   bsz  = boolsizeof(bool:pair);
   /* Float vectors[3] */
   x0  = palloc(pair, vsz);  x1 = palloc(pair, vsz);
   x0j = palloc(pair, vsz);
   v0  = palloc(pair, vsz);  v1 = palloc(pair, vsz);
   ac  = palloc(pair, vsz);  ax = palloc(pair, vsz);
   wv  = palloc(pair, vsz);
   /* xm0, xm1, acm, mmj must be consecutive */
   xm0 = palloc(pair, cvsz);
   xm1 = &xm0[3];  acm = &xm0[6];  mmj = &xm0[9];
   /* Floats */
   m  = palloc(pair, fsz);  mi = palloc(pair, fsz);  mj = palloc(pair, fsz);
   w  = palloc(pair, fsz);
   /* Ints and bools */
   i  = palloc(pair, isz);  j = palloc(pair, isz);  k = palloc(pair, isz);
   diag = palloc(pair, bsz);  act = palloc(pair, bsz);
 }

/* Pairs are represented symmetrically.	 Setup initializes these.	*/

void setup()
 { with (pair)
    { /* Initialize coordinates */
      *i = pcoord(0);  *j = pcoord(1);  *k = pcoord(2);
      /* Identify active processors */
      *act = (*i != *j);  *diag = (*i == *j);
    }
 }

FILE *datap;      /* Stream for reading data */

/* Load inital positions, velocities, also masses, gravitational const. */
/* Leaves data for Pu in processors [i][0][k] where i*NS + k = u.	*/

void init()
 { int u,i,k;  float a, b, c;
   with (pair)
   {for ( u = 0; u < NB; u++ )
     { fscanf(datap, "%e%e%e", &a, &b, &c);  i = u/NS;  k = u%NS;
       [i][0][k]x0[0] = a;  [i][0][k]x0[1] = b;  [i][0][k]x0[2] = c;
     };
    for ( u = 0; u < NB; u++ )
     { fscanf(datap, "%e%e%e", &a, &b, &c);  i = u/NS;  k = u%NS;
       [i][0][k]v0[0] = a;  [i][0][k]v0[1] = b;  [i][0][k]v0[2] = c;
     };
    for ( u = 0; u < NB; u++ )
      { fscanf(datap, "%e", &a);  i = u/NS;  k = u%NS;  [i][0][k]*m = a; };
    fscanf(datap, "%e", &G);
    /* Spread data across rows */
    x0[0] = copy_spread(&x0[0], 1, 0);
    x0[1] = copy_spread(&x0[1], 1, 0);
    x0[2] = copy_spread(&x0[2], 1, 0);
    v0[0] = copy_spread(&v0[0], 1, 0);
    v0[1] = copy_spread(&v0[1], 1, 0);
    v0[2] = copy_spread(&v0[2], 1, 0);
    *m  = copy_spread(m, 1, 0);
   }
 }

/* Insert na asteroids at radius ar with total mass am.  This assumes	*/
/* that an orbit of radius 1 about origin corresponds to velocity 2*pi	*/
/* (e.g., au-earth-yr units).  Uses parallel trig.			*/

void mkast(int na, float ar, float am)
 { float:pair sx, cx, arg;
   int:pair pn;  /* particle number */
   float spd;  int nn;
   nn = NB + na;
   spd = TWOPI/sqrt(ar);
   printf("Making %4d asteroids, r = %9.5f, spd = %9.5f, tot mass = %9.2e\n",
		  na,		     ar,	  spd,		    am);
   with (pair) 
   { pn = *i*NS + *k;  /* Particle represented by this processor */
     where ( (NB <= pn) & (pn < nn) )
      { arg = (TWOPI*(pn - NB))/na;
        sx = sin(arg);  cx = cos(arg);
        x0[0] =  cx*ar;  v0[0] = (-sx)*spd;
        x0[1] =  sx*ar;  v0[1] =  cx*spd;
        x0[2] = 0.0;     v0[2] = 0.0;
        *m = am/na;
      }}  /* end where, end with (pair) */
   NB = nn;
 }


/* Print current state of system according to PX.  See comments for	*/
/* main program for PX settings.					*/

void print_state()
 { int u, i, k, np;
   if ( (PX >= 0) && (PX < NB) ) np = PX; else np = NB;
   if ( np != 0) /* print headings */
   printf(
  "\n       x         y         z         vx        vy        vz        m\n");
   for ( u = 0; u < np; u++ )
    { i = u/NS;  k = u%NS;
      printf("%2d %9.5f %9.5f %9.5f %9.5f %9.5f %9.5f %9.2e\n",
	    u, [i][i][k]x0[0], [i][i][k]x0[1], [i][i][k]x0[2],
               [i][i][k]v0[0], [i][i][k]v0[1], [i][i][k]v0[2], [i][i][k]*m); }
 }

/* Print x-components of xx in nx by nx pipes (for testing). */

void print_x(char lbl[], float:current *xx, int nx)
 { int u, v, p;
   printf("\n%s\n", lbl); 
   for ( u = 0; u < nx; u++)
    { for ( v = 0; v < nx; v++)
       for ( p = 0; p < NS; p++)
        { if ( p%8 == 0 ) printf("\n");
          printf("%8.1e ", [u][v][p]*xx);
        }
      printf("\n");
    }
 }


/* Spread xv[3] along i-axis (0-axis) from diagonal, leaving result in dv, */
/* Called with all processors active and leaves all active.  Uses wv.	   */

void spreadi(float:pair *dv, float:pair *xv)
 { int u;
   for ( u = 0;  u < 3;  u++ )
    { where (*diag) wv[u] = xv[u]; else wv[u] = -FLT_MAX;
      dv[u] = spread(wv[u], 0, CMC_combiner_max);
    }
 }

/* Sum accelerations, leaving results in all processors of each row.	*/
/* Called with all processors active.					*/

void sumac()
 { /* Return moving ac's to original positions */
  everywhere
  {to_torus_dim(acm, acm, vsz, 2, SS);
   /* print_x("acm after return", acm, 2); 	For testing */
   /* Accumulate moving ac's along i axis */
   acm[0] = spread(acm[0], 0, CMC_combiner_add);
   acm[1] = spread(acm[1], 0, CMC_combiner_add);
   acm[2] = spread(acm[2], 0, CMC_combiner_add);
   /* Combine on diagonal */
   where (*diag) { ac[0] += acm[0];  ac[1] += acm[1];  ac[2] += acm[2]; }
   /* Sum accelerations along j-axis */
   ac[0] = spread(ac[0], 1, CMC_combiner_add);
   ac[1] = spread(ac[1], 1, CMC_combiner_add);
   ac[2] = spread(ac[2], 1, CMC_combiner_add);
   /* print_x("ac after accumulation", ac, 2);	For testing */
  } /* end everywhere */
 }


/* Calculate accelerations.  x0j, mj already have "0" positions and	*/
/* scaled masses for "j" particles.  Called with all processors active.	*/

void accel()
 { float:pair d0[3], d1[3], r0, r1, w, f;  int u, s;
  everywhere
  {spreadi(xm1, x1);		      /* Spread x1 along i-axis into xm1 */
   /* Now xm1 = [j][j][k]x1,  x1 = [i][i][k]x1 */
   /* print_x("accel - xm1 after spreadi", xm1, 2);	For testing */
   *mmj = *mj;				/* Load conveyer */
   r0 = r1 = 0.0f;			/* To accumulate sum of squares */
   for ( u = 0;  u < 3;  u++ )	      /* First acceleration calculation */
    { ac[u] = acm[u] = 0.0f;   xm0[u] = x0j[u];	       /* Load conveyor */
      d0[u] = xm0[u] - x0[u];  r0 += d0[u]*d0[u];
      d1[u] = xm1[u] - x1[u];  r1 += d1[u]*d1[u]; }
   r0 = sqrt(r0);  r1 = sqrt(r1);		/* Euclidean distances */
   /* print_x("accel - r1", &r1, 2);	For testing */
   w = *mmj/(r0*r1*(r0 + r1));			/* Multiplier */
   where (*act)					/* Form acceleration */
    { for ( u = 0;  u < 3;  u++ ) ac[u] = w*(d0[u] + d1[u]); }
   /* print_x("accel - ac after step 0", ac, 2);	For testing */
   /* Loop for intermediate (systolic) calculations */
   for ( s = 1;  s < SS;  s++)
    { to_torus_dim(xm0, xm0, cvsz, 2, 1);		/* Advance conveyor */
      r0 = r1 = 0.0f;
      for ( u = 0;  u < 3;  u++ )
       { d0[u] = xm0[u] - x0[u];  r0 += d0[u]*d0[u];
	 d1[u] = xm1[u] - x1[u];  r1 += d1[u]*d1[u];
         wv[u] = d0[u] + d1[u]; }
      r0 = sqrt(r0);  r1 = sqrt(r1);		/* Euclidean distances */
      w = r0*r1*(r0 + r1);
      f = *mi/w;  w = *mmj/w; 	 	/* Multipliers */
      for ( u = 0;  u < 3;  u++ )		/* Form accelerations */
       { ac[u] += w*wv[u];  acm[u] -= f*wv[u]; }
      /* print_x("accel - ac after step", ac, 2);	For testing */
      /* print_x("accel - acm after step", acm, 2);	For testing */
    } /* End systolic loop */
   to_torus_dim(xm0, xm0, cvsz, 2, 1);		/* Advance conveyor */
   r0 = r1 = 0.0f;
   for ( u = 0;  u < 3;  u++ )
    { d0[u] = xm0[u] - x0[u];  r0 += d0[u]*d0[u];
      d1[u] = xm1[u] - x1[u];  r1 += d1[u]*d1[u];
      wv[u] = d0[u] + d1[u]; }
   r0 = sqrt(r0);  r1 = sqrt(r1);		/* Euclidean distances */
   w = *mmj/(r0*r1*(r0 + r1));			/* Multiplier */
   for ( u = 0;  u < 3;  u++ ) ac[u] += w*wv[u];	/* Accelerations */
   /* print_x("accel - final ac", ac, 2);		For testing */
   /* print_x("accel - final acm", acm, 2);		For testing */
   sumac();					/* Add them up */
  } /* end everywhere */
 }


/* Initialize masses, x1, and v1 for green1.	*/

void initgreen(float dt)
 { int u;
   with (pair)
    { *mi = G**m;		/* mi is scaled mass of particle i */
      /* Spread scaled masses for particle j */
      *mj = [*j][*j][*k]*mi;
      /* Calculate initial prediction for x1, v1 */
      for ( u = 0; u < 3; u++ )
	{ v1[u] = v0[u];  x1[u] = x0[u] + dt*v0[u]; }
    }  /* End with (pair) */
 }

/* Greenspan's 1st-order method.  Performs ns steps of time dt		*/
/* with solution tolerance eps.  Returns number of evaluations.		*/
/* Leaves number of iterations which were stopped at limit in cfail.	*/
/* x1, v1 must be initialized to some reasonable prediction.  Leaves	*/
/* prediction for next time step in x1, v1.				*/
/* Some printing for debugging is provided in comments.			*/

int cfail;   /* Counts number of iterations that failed to converge */

int green1(int ns, float dt, float eps)
 { int k, ic, v, lim=10, u;  float mx, hdt;  float:pair nx, px;
   v = 0;  hdt = 0.5f*dt;
   with (pair) everywhere
   {  for (k = 0; k < ns; k++)
       { ic = 0;
	 spreadi(x0j, x0);			/* x0 for moving particles */
	 /* This is a repeat-until loop: */
   	 do
          { accel();  v++;  ic++;
	    px = 0.0f;	/* For max change */
	    for ( u = 0;  u < 3;  u++ )
	       { nx = v0[u] + dt*ac[u];			/* New velocity	*/
		 px >?= fabs(nx - v1[u]);		/* Max change	*/
		 v1[u] = nx;
		 nx = x0[u] + hdt*(v0[u] + v1[u]);	/* New position	*/
		 px >?= fabs(nx - x1[u]);		/* Max change	*/
		 x1[u] = nx; 
	       } /* end for u */
     /* print_x("green1, new x1", x1, 2);	For testing */
     /* print_x("green1, new v1", v1, 2);	For testing */
	      mx = (>?= px);				/* Global max */
	    if ( (mx < eps) || ( ic == lim ) ) break;
	  }
	 while (1);
	 /* Update x0, v0 and predict next x1, v1 */
	 for ( u = 0;  u < 3;  u++ )
	  { v0[u] = v1[u];  v1[u] += dt*ac[u];
	    x0[u] = x1[u];  x1[u] += hdt*(v0[u] + v1[u]);
	  }
	 if (ic == lim) cfail++;
       } ;   /* end for */
    }  /* end  with (pair) */
   return(v);
 }

/* Main program:  reads problem and control information from data file	*/
/* named by user and performs calculation.				*/
/* Print control:							*/
/*	  0 - just performance data					*/
/*	< 0 - all							*/
/*	> 0 - number given						*/

void main()
 { int ns, np, k, vl, nx;  float dt, eps, ar, am;
   char fnm[50];
   /* Initialize the pair shape with suitable geometry */
   printf("Enter NM, NS for [NM][NM][NS] shape> ");
   scanf("%d%d", &NM, &NS);
   allocate_shape(&pair, 3, NM, NM, NS);
   SS = NS/2;  nx = NM*NS;	/* nx = eventual NB */
   create_globals();
   /* Loop for test cases */
   while ( printf("Data file (or q)> "),
	   scanf("%s", fnm),
	   ( fnm[0] != 'q' ) || ( fnm[1] > 0 ) )
    { printf("Print ctl> ");  scanf("%d", &PX);
      /* Performs np integrations, each ns steps of dt */
      datap = fopen(fnm, "r");
      fscanf(datap, "%d", &NB);
      setup();   /* Setup geometric structure */
      printf("Initializing, %d explicit bodies.\n", NB);
      init();    /* Read data for individual bodies */
      /* Read "asteroid" descriptions */
      while ( fscanf(datap, "%d", &ns), ns > 0)
       { fscanf(datap, "%e%e", &ar, &am);  mkast(ns, ar, am); }
      if ( NB < nx ) mkast(nx-NB, 3.113, 0.01);	/* Fill with extra asteroids */
      fscanf(datap, "%d%d%e%e", &ns, &np, &dt, &eps);
      fclose(datap);
      printf("\n   NB = %d,  Geometry is [%d][%d][%d]",
		        NB,               NM, NM, NS   );
      printf("\n   G = %13.6e  delta-t = %12.5e  epsilon = %12.5e\n",
			  G,		   dt,		     eps);
      printf("\n simulated time = 0");
      print_state();
      initgreen(dt);	/* Distribute scaled masses, etc. */
      for ( k = 0; k < np; k++ )
       { CM_timer_clear(1);  CM_timer_start(1);
         cfail = 0;
         vl = green1(ns, dt, eps);
         CM_timer_stop(1);  CM_timer_print(1);
         printf(
        "\n simulated time = %9.3f  %d evaluations  %d iterations limited\n",
	             (k + 1)*ns*dt, vl,            cfail);
         print_state();
       };  /* end for */
    }  /* end while */
 }

/*********************************************************************
  Data files:  as for nb3dg2.cs
*************************************************************************/
